#include "option.h"
#include "stock.h"
#include "haar.h"
#include "falgorithm.h"
#include <cmath>
#include <iostream>
#include <vector>
using namespace std;

Option::Option(Stock theStock,double theTime,double strikePrice,double theAmount){
  stock = theStock;
  etime = theTime;
  P = strikePrice;
  amount = theAmount;
}

Option::Option(){
  Stock theStock;
  Option(theStock,0,0,0);
}

// BEGIN code1
vector<double> Option::holdingPortfolio(Path path){
  vector<double> hportfolio;
  for(int i = 0;i < path.times.size(); i++){
    if(path.times[i] < etime)
      hportfolio.push_back(value(path.prices[i],path.times[i]));
    else
      hportfolio.push_back(evalue(path.prices[i]));
  }
  return hportfolio;
}
// END code1

// BEGIN code2
double Option::get_d1(double stockPrice,double time)
{
  return (log(stockPrice/P) + (stock.r + pow(stock.sigma,2.0)/2.0) * (etime - time)) 
    / (stock.sigma * sqrt(etime - time));
}
// END code2

// BEGIN code3
double Option::get_d2(double d1,double time)
{
  return d1 - stock.sigma * sqrt(etime - time);
}
// END code3

//european calls
EuCall::EuCall(Stock theStock,double theTime,double strikePrice,double amount)
  :Option( theStock,theTime,strikePrice,amount){
  //call constructor of super class, empty body
}
EuCall::EuCall()
  :Option(){
}

double EuCall::get_phi(double d1)
{
  return amount * cumulative(d1);
}

double EuCall::exactPrice(){
  double d1 = get_d1(stock.St0, 0.0);
  double d2 = get_d2(d1, 0.0);

  double optionPrice = stock.St0 * cumulative( d1 ) - 
    P * exp( -stock.r * etime ) * cumulative( d2 );
  return optionPrice;
}

// BEGIN code4
double EuCall::value(double stockPrice,double time){
  double d1 = get_d1(stockPrice,time);
  double d2 = get_d2(d1,time);
  double value = amount * (stockPrice * cumulative(d1) 
			   - P * exp( -stock.r * (etime - time)) * cumulative(d2));
  return value;
}
// END code4

// BEGIN code5
double EuCall::evalue(double stockPrice){
  return amount * max(0.0,(stockPrice - P));
}
// END code5

//t_BS must be a subset of the times in the path object
//the last element of t_BS is same as the expiration time of the call option
//(t0,t1,t2,....tBS)
vector<double> EuCall::BSPortfolio(Path path){
  double BSP = value(path.prices[0], 0.0);
  vector<double> BSPs;
  BSPs.push_back(BSP);

  double deltaT=path.times[0];
  double d1,phi,xi;
  for(int i=0; i < path.times.size()-1; i++)
    {
      d1 = get_d1(path.prices[i],path.times[i]);
      phi = amount * cumulative(d1);
      xi = BSP - phi * path.prices[i];
      BSP = xi * exp(stock.r * deltaT) + phi * path.prices[i+1];
      BSPs.push_back(BSP);
      deltaT = path.times[i+1] - path.times[i];
    }
  return BSPs;
}

// BEGIN code6
vector<double> EuCall::HHPayoff(int j_H,double stockPrice,double time,double deltaT){
  if(time >= etime){
    cout<<"Hedging time should not greater or equal to option's expiration time"<<endl;
    exit(1);
  }
    
  vector<double> underlying = stock.oneStepUnderlying1(j_H,stockPrice,deltaT);
  vector<double> Xis;
  double Xi;
  //  cout<< abs((time + deltaT) - etime) <<endl;
  if(  abs((time + deltaT) - etime) < 0.001 ){
    for(int j=0;j<underlying.size();j++){
      Xi = amount * max(0.0,underlying[j] - P);
      Xis.push_back(Xi);
    }
  }
  else{
    for(int j=0;j<underlying.size();j++){
       Xi = value(underlying[j],time+deltaT);
      Xis.push_back(Xi);
    }
  }
  return Xis;
}
// END code6

//european put
double EuPut::exactPrice(){
  double d1 = get_d1(stock.St0, 0.0);
  double d2 = get_d2(d1, 0.0);
  double optionPrice = P * exp( -stock.r * etime ) * cumulative( -d2 ) 
    - stock.St0 * cumulative( -d1 );
  return optionPrice;
}

double EuPut::value(double stockPrice,double time){
  double d1 = get_d1(stockPrice,time);
  double d2 = get_d2(d1,time);
  double value = amount * (P * exp(-stock.r * (etime - time)) * cumulative(-d2) 
			   - stockPrice * cumulative(-d1));
  return value;
}

double EuPut::evalue(double stockPrice){
  return amount * max(0.0,(P - stockPrice));
}

vector<double> EuPut::BSPortfolio(Path path){
  double BSP = value(path.prices[0], 0.0);
  vector<double> BSPs;
  BSPs.push_back(BSP);

  double deltaT=path.times[0];
  double d1,phi,xi;
  for(int i=0; i < path.times.size()-1; i++)
    {
      d1 = get_d1(path.prices[i],path.times[i]);
      phi = -amount * cumulative(-d1);
      xi = BSP - phi * path.prices[i];
      BSP = xi * exp(stock.r * deltaT) + phi * path.prices[i+1];
      BSPs.push_back(BSP);
      deltaT = path.times[i+1] - path.times[i];
    }
  return BSPs;
}

vector<double> EuPut::HHPayoff(int j_H,double stockPrice,double time,double deltaT){
  if(time >= etime){
    cout<<"Hedging time should not greater or equal to option's expiration time"<<endl;
    exit(1);
  }

  vector<double> underlying = stock.oneStepUnderlying1(j_H,stockPrice,deltaT);

  vector<double> Xis;
  double Xi;
  if( abs((time + deltaT) - etime) < 0.001 ){
    for(int j=0;j<underlying.size();j++){
      Xi = amount * max(0.0,P - underlying[j]);
      Xis.push_back(Xi);
    }
  }
  else{
    for(int j=0;j<underlying.size();j++){
      Xi = value(underlying[j],time+deltaT);
      Xis.push_back(Xi);
    }
  }
  return Xis;
}

EuPut::EuPut(Stock theStock,double theTime,double strikePrice,double amount)
  :Option( theStock,theTime,strikePrice,amount){
}
EuPut::EuPut()
  :Option(){
}
double EuPut::get_phi(double d1)
{
  return -amount * cumulative(-d1);
}
